/*
    $Id: astrotime.h 53 2009-11-01 18:03:28Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _ASTROTIME_H
#define _ASTROTIME_H

#include <ctime>
#include <sys/time.h>

#include <string>
#include <iosfwd>

template<class T>  T julian( int year, int month, T day ) {
    T result = (3057*month)/100 + (365*year) + 1721028;
    result = result + (year/4 - year/100 + year/400 + day);
    if( month > 2 ) return( result);
    int j = 2;
    if( year%4 == 0 )   j -= 1;
    if( year%100 == 0 ) j -= 1;
    if( year%400 == 0 ) j += 1;
    result += j;
    return( result );
}

template<class T>  double s0( T jd ) {
    const double c3 = 1.0775925925925926E-06;
    const double c2 = 2.7379093507870371E-03;
    const double c1 = 0.2790572732638889;
    double d = (double)jd - 2451545.5;
    double t2 = d/36525.0;
    double result = c1 + c2*d + c3*t2*t2;
    return result - floor( result );
}

struct Date {
    int year;
    int month;
    int day;
    operator int () const { return julian( year, month, day ); }
    Date(int y = 2000, int m = 1, int d = 1) : year(y), month(m), day(d) {}
    Date(double jd) {
        const  time_t sstart = 86400*(int)(jd-julian( 1970, 1, 1 ));
        struct tm broken;
        gmtime_r( &sstart, &broken );
        year = broken.tm_year + 1900;
        month = broken.tm_mon + 1;
        day = broken.tm_mday;
    }
};

struct Time {
    int hour;
    int min;
    int sec;
    operator double () const { return hour+min/60.0+sec/3600.0; }
    Time( int h = 0, int m = 0, int s = 0 ) : hour(h), min(m), sec(s) {}
    Time( double pd ) : hour( (int)pd ), min( (int)fmod( 60*pd, 60 ) ), sec( (int)fmod( fmod( 3600*pd, 3600 ), 60 ) ) {}
};

namespace std {
    std::istream& operator>> ( std::istream&, Date& );
    std::istream& operator>> ( std::istream&, Time& );
    std::ostream& operator<< ( std::ostream&, const Date& );
    std::ostream& operator<< ( std::ostream&, const Time& );
};

enum {ASTRO, CONV};

/** \brief        (UTC). */
class AstroDateTime {
    bool realt;
    int UTC_msec;
    int at_julian;
    double at_s0;
    struct timeval hightime;
    time_t  start_midnight;
    struct tm UTC;
    struct tm UTstart;
    struct tm LT;
public:
    AstroDateTime();                        ///<   
    AstroDateTime( int, int, int );         ///<   
    ~AstroDateTime() {}
    void current( int ovmid = 0 );          ///<   
    void current( int h, int m, double s ); ///<  - 
    void local();                           ///<    
    int year() { return UTC.tm_year; }      ///<  4-  
    int month() { return UTC.tm_mon; }      ///<   
    int day() { return UTC.tm_mday; }       ///<   
    int hour() { return UTC.tm_hour; }      ///<   
    int min() { return UTC.tm_min; }        ///<   
    int sec() { return UTC.tm_sec; }        ///<   
    int msec() { return UTC_msec; }         ///<   
    int diffUT() { return UTC.tm_gmtoff/60/60; }
    int julian( int Y, int m, int d );      ///<       \a Y - \a m - \a d
    int julian() { return at_julian; }      ///<      
    std::string strTime( int pr = 3 );      ///<     HH:MM:SS.SSS
    std::string strDate();                  ///<   \b     YYYY-MM-DD
    std::string strFileName();              ///<   \b     YYMMDD
    double s0( int );                       ///<        JD,   
    double s0(){ return s0( at_julian ); }  ///<         ,   
    double currentSec();                    ///<        
    double sidTime( double longitude);      ///<    ( longitude   )
    double rsidTime( double longitude);     ///<    ( longitude  )
};

#endif // _ASTROTIME_H
