/*
    $Id: file_input.cpp 53 2009-11-01 18:03:28Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cassert>
#include <stdexcept>
#include <sstream>
#include <iostream>

#include <boost/io/ios_state.hpp>
#include <boost/numeric/ublas/vector.hpp>

#include "atmos.h"
#include "file_input.h"

const std::string::size_type file_input::payload_col = 22;

file_input::file_input() : m_line(0) {
}
file_input::~file_input() {
}
void file_input::operator() ( std::istream& stm ) {
	do_parse( stm );
}
file_input::size_type file_input::line() const throw() {
	return m_line;
}
//FIXME: use boost string_algo
std::string file_input::strip_spaces( const std::string& str ) {
	std::string ret(str);
	boost::trim_if( ret, boost::is_space() );
	return ret;
}
void file_input::do_parse( std::istream& stm ) {
	std::string line;
	std::string::value_type type;

	boost::io::ios_exception_saver esaver( stm );
	stm.exceptions(std::ios_base::badbit | std::ios_base::failbit);

	Date cdate;
	Time ctime;
	ublas::vector<double> mean( atmos::channels );
	ublas::vector<double> vars( atmos::indixes );
	ublas::vector<double> rho1( atmos::indixes );
	ublas::vector<double> rho2( atmos::channels );
	try {
	while( std::getline( stm, line ) ) {
		if( line.size() == 0 ) continue; // skip empty line
		type = line.at(0); // detect type
		if( type == '#' ) { // comment
			on_comment( line );
		} else if ( type == 'P' ) {
			std::string::size_type eq = line.find_first_of("=");
			if( eq == std::string::npos ) throw failure("Malformed Parameter line", m_line);
			on_parameter( strip_spaces( line.substr( payload_col, eq-payload_col ) ), strip_spaces( line.substr( eq+1 ) ) );
		} else if ( type == 'O' ) {
			std::istringstream iss( line.substr( payload_col ) );
			object obj;
			iss.exceptions(std::ios_base::badbit | std::ios_base::failbit);
			try {
				iss >> obj;
			} catch( const std::ios::failure& e ) {
				throw failure("Malformed Object line", m_line);
			}
			on_object( obj );
		} else if ( type == 'M' ) {
			std::istringstream iss( line );
			iss.exceptions(std::ios_base::badbit | std::ios_base::failbit);
			try {
				iss.ignore(); // ignore M
				iss >> cdate >> ctime;
			} catch( const std::ios::failure& e ) {
				throw failure("Malformed Mode line", m_line);
			}
			if( line.substr( payload_col ).find("Background") != std::string::npos ){
				on_background( cdate, ctime );
			} else if ( line.substr( payload_col ).find("Normal") != std::string::npos ) {
				on_normal( cdate, ctime );
			} else if ( line.substr( payload_col ).find("Flux") != std::string::npos ) {
				on_flux( cdate, ctime );
			}
		} else if ( type == 'm' ) {
			std::istringstream iss( line );
			iss.exceptions(std::ios_base::badbit | std::ios_base::failbit);
			try {
				iss.ignore(); // ignore m
				iss >> cdate >> ctime >> mean >> vars >> rho1 >> rho2;
			} catch( const std::ios::failure& e ) {
				throw failure("Malformed measurements line", m_line);
			}
			on_measurements( cdate, ctime, mean, vars, rho1, rho2 );
		} else { // unknown

		}
		m_line++;
	}
	} catch( const std::ios::failure& e ){
		if( !stm.eof() ) throw; // eofbit is always set with failbit
	}
}

file_input::failure::failure( const char* what, size_type line ):
	std::runtime_error( what ), m_line(line) {
}
file_input::size_type file_input::failure::line() const {
	return m_line;
}

std::istream& std::operator>>( std::istream& stm, file_input::object& obj ){
	//O 2009-07-19 18:04:29 5191 Eta_UMa 13 47 32 +49 18 48  1.86 -0.19 B55
	double v,bv;
	return stm >> obj.hd >> obj.name >> obj.coordinates >> v >> bv >> obj.sclass;
}
