/*
    $Id: indices.cpp 53 2009-11-01 18:03:28Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <math.h>
#include <algorithm>
#include <numeric>

#include <boost/format.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/io/ios_state.hpp>

#include "atmos.h"
#include "indices.h"

const unsigned int nchan = atmos::channels;
const unsigned int ninds = atmos::indixes;                // Restoration: based on fixed number of scint.indices:

const double min_flux = 2.0;

indices::indices():
	microexposure( 1.0 ),
	deadtime( scalar_vector<double>( nchan, 20e-6 ) ),
	nonpois( scalar_vector<double>( nchan, 1.0 ) ),
	scatter( scalar_vector<double>( nchan, 0.0 ) ),
	backgrd( scalar_vector<double>( nchan, 0.0 ) ),
	flux(0,nchan), desi(0,nchan), inds(0,ninds) {
}
indices::~indices(){
}

void indices::calculation( const vector<double>& cmean, const vector<double>& cvars, const vector<double>& crho1, const vector<double>& crho2 ) {
	assert( cmean.size() == nchan );
	assert( cvars.size() == ninds );
	assert( crho1.size() == ninds );
	assert( crho2.size() == nchan );
                                // Input filtering for star escape
                                // Re-normalization
    vector<double> renorm = mass_order_prod( cmean );
    vector<double> vars = element_prod( cvars, renorm );
    vector<double> rho1 = element_prod( crho1, renorm );
    vector<double> rho2 = element_prod( crho2, subrange( renorm, 0, nchan ) );
                                // Poisson and nonlinearity correction
    vector<double> tauxi = element_prod( cmean, deadtime )/( -1.0*microexposure );
    vector<double> wtxi  = apply_to_all( tauxi, functor::lambertW<double>() );
    subrange( vars, 0, nchan ) -= element_prod( element_prod( cmean, (scalar_vector<double>( nchan, 1.0 ) + tauxi*2.0) ), nonpois );
    vector<double> mean = element_prod( cmean, apply_to_all( wtxi*(-1.0), functor::exp<double>() ) );
    mean = element_prod( ( scalar_vector<double>( nchan, 1.0 ) - scatter), mean )- backgrd*microexposure;
    if( *min_element( mean.begin(), mean.end() ) < min_flux )
	throw failure( (boost::format("Non star's data %1%") % mean ).str() );
    matrix_push_row( flux, mean );
    vector<double> corr = apply_to_all( wtxi, functor::exp<double>() ) + tauxi;
    corr = apply_to_all( corr, functor::inv<double>() );
    vector<double> factor = mass_order_prod( corr );
    vars = element_prod( vars, factor );
    rho1 = element_prod( rho1, factor );
    rho2 = element_prod( rho2, subrange( factor, 0, nchan ) );
                                // Indices calculation
    corr = apply_to_all( mean, functor::inv<double>() );
    factor = mass_order_prod( corr );
    vars = element_prod( vars*1.25 - rho1*0.25, factor );
    vector<double> sigm =  subrange( vars, 0, nchan );
    subrange( vars, nchan, ninds ) =  subrange( mass_order_summ( sigm ), nchan, ninds ) - subrange( vars, nchan, ninds )*2.0;
    matrix_push_row( inds, vars );
                                // DESI calculation
    matrix_push_row( desi, ( sigm*3.0 +  element_prod( rho2 - subrange( rho1, 0, nchan )*4.0 , subrange( factor, 0, nchan ) ) )/4.5 );
}

void indices::reset() throw() {
    flux.resize(0,nchan);
    inds.resize(0,ninds);
    desi.resize(0,nchan);
}

vector<datum> indices::result( const matrix< double >& inp ) const {
    vector<datum> out;
    unsigned int nrow = inp.size1();
    unsigned int ncol = inp.size2();
    out.resize( ncol );
    double n = (double)nrow;

    if( nrow < 3 ){
	    throw failure( (boost::format("Too few measurements: %1%") % nrow ).str() );
    }

    for( unsigned int i = 0; i < ncol; i++ ) out(i).mean = sum(column( inp,i ))/n;
    for( unsigned int i = 0; i < ncol; i++) {
        double vars = inner_prod( column( inp, i ), column( inp, i ))/n - out(i).mean*out(i).mean;
        out(i).errs = sqrt( vars/(n-1) )/out(i).mean;
        if( isnan(out(i).errs) || isinf(out(i).errs)  ) {
            //FIXME:
            out(i).errs = 0.0;
        }
    }
    return out;
}

void indices::set_background( const vector<double>& vals  ) {
    assert( backgrd.size() == vals.size() );
    backgrd = vals;
}

void indices::set_microexposure( double mex ) {
    microexposure = mex;
}

void indices::set_counters( const vector<double>& nl, const vector<double>& np, const vector<double>& sc ) {
    assert( deadtime.size() == nl.size() );
    assert( nonpois.size() == np.size() );
    assert( scatter.size() == sc.size() );

    deadtime = nl*1e-6;
    nonpois = np;
    scatter = sc;
}
