/*
    $Id: integrals.cpp 53 2009-11-01 18:03:28Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cmath>
#include <algorithm>
#include <numeric>
#include <boost/format.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/symmetric.hpp>
#include <lsp/singular_decomposition.h>

#include "typesdef.h"
#include "indices.h"
#include "integrals.h"

const double KM2M = 1000;
const double POWSEE = 0 ;       // Power of Cn2 moment for seeing
const double POWISP = 5.0/3.0;  // Power of Cn2 moment for isoplanatic angle
const double POWEFF = 1.0;      // Power of Cn2 moment for Effective altitude
const double POWISK = 2.0;      // Cn2 second moment power

const double MINRELW = 6e-4 ;   // Minimal ratio of the singular value to the maximal one

static const double powers[] = { POWSEE, POWISP, POWEFF, POWISK };

const unsigned int npower = sizeof(powers)/sizeof(powers[0]);

static const double HBOUND = 0.38;// Height of the boundary layer [kilometers] to fit the Cn2-integral
static const int TOPLAYERS =  4; // Number of rejected high-altitude layers
const double JPOW_SEE = 0.6 ;
const double LPOW_SEE = -0.2;
const double KS = 1.73588e+07;  // Calibration constant in seeing formula (AstL 45, p.399):
                                //\f$ seeing[''] = KS * \lambda_{eff}^{-1/5} M_0^{3/5} *
                                //\sec^{-1}(z)\f$, where \em lambda_eff is in [mkm]*/

const double JPOW_ISP = -0.6;   // Power of Cn-5/3th moment for isoplan. angle
const double LPOW_ISP = 1.2;    // Power of lambda for isoplan. angle
const double KP = 0.000756348;  // Calibration constant in isoplanatic angle formula (AstL 45, p.399)
                                // \f$ \theta_0[''] = KP * (M_{5/3} /\lambda_{eff}^2)^{-3/5} * \sec^{8/5}(z) \f$, where
                                // \f$ KP = (2.905*(2\pi)^2*(MKM2M)^{-2})^{-3/5}*206265 \f$*/
const double KT = 0.175;        // Atmospheric time constant empirical calibration constant for DESI for 1,3ms
                                // integrations in diameter=2cm apertures with -1km shift

const double SPOW_TC  = -0.6;   // Power of DESI for atm. time constant

template<class T> vector<T>& reject_elements( vector<T>& x, T minval ) {
    for( typename vector<T>::iterator it = x.begin(); it != x.end(); ++it ){
        *it = ( *it > minval )? *it : T(0);
    }
    return x;
}

class to_seeing {
    double secz;
    double leff;
public:
    to_seeing( double s, double l ) : secz(s), leff(l) {}
    double operator()( double& em ) {
        return KS*pow( em/secz, JPOW_SEE )*pow( leff, LPOW_SEE );
    }
};

class to_isopla {
    double secz;
    double leff;
public:
    to_isopla( double s, double l ) : secz(s), leff(l) {}
    double operator()( double& em ) {
        double zpow = em/pow( secz, POWISP+1 );
        return KP*pow( zpow, JPOW_ISP )*pow( leff, LPOW_ISP );
    }
};

class to_altitude {
    double secz;
public:
    to_altitude( double s ) : secz(s) {}
    double operator()( double& em1, double& em2 ) {
        double zpow = pow( secz, POWISP-POWEFF );
        return 0.001*pow( em1/em2/zpow, 1.0/( POWISP-POWEFF ) );
    }
};

class to_timeconst {
    double secz;
public:
    to_timeconst( double s ) : secz(s) {}
    double operator()( double em ) {
        return KT*pow( em, SPOW_TC );
    }
};

bool Unapt( const double& em ) {
        return isnan( em )||isinf( em )||( em < 0.0 );           // Phisical restriction !!!
}

integrals::integrals():
	m_lambda_eff( 0 ),
	m_nscind( 0 ) {
}
integrals::~integrals(){
}

//FIXME:
datum integrals::result( const vector<double>& y ) {
    vector<double> x( y.size() );
    x.resize( std::remove_copy_if( y.begin(), y.end(), x.begin(), Unapt ) - x.begin());
    double n = (double)x.size();
    datum f;
    if( n < 2 ) return f;
    f.mean = sum( x )/n;
    f.errs = sqrt((inner_prod( x, x )/n - f.mean*f.mean)/(n-1.0))/f.mean;
    if( f.errs < 0 || isnan(f.errs) || isinf(f.errs)  ) {
	//FIXME: must be checked outside the function or throw
        f.errs = 0.0;
    }
    return  f;
}

void integrals::coefficients( const std::string& spectral_type, const Weights& awf ) {
    if( spectral_type == m_spectrum ) return;
    m_nscind = awf.weight_functions_number();
    m_lambda_eff = awf.lambda_eff();
    unsigned int i = 0;
    vector<double> zgrid = awf.grid();
    while( zgrid( i ) < HBOUND ) i++;
    vector<double> h = subrange( zgrid, i, zgrid.size()-TOPLAYERS );
    Weights wfs( awf.apertures_number(), h );
    wfs.interpolate( awf ) ;
    m_dcf.resize( npower, m_nscind );
    for( unsigned int ipow = 0; ipow < npower ; ipow++ ) {
        double power = -powers[ipow]/2.0;
        double scale = pow( KM2M, powers[ipow] );
        vector<double> c = apply_to_all( apply_to_all( h, functor::log<double>() )*power, functor::exp<double>() );
        vector<double> b = apply_to_all( c, functor::inv<double>() )*scale;
        matrix<double> A( h.size(), m_nscind );            // Make the matrices of weights
        for( unsigned int j = 0; j < m_nscind; j++ ) {
            column( A, j ) = element_prod( row( wfs.weights(), j ), c );
        }
        matrix< double > left  = identity_matrix<double>( A.size1() );
        matrix< double > right = identity_matrix<double>( A.size2() );
        lsp::singular_decomposition< matrix< double > > sd( A );
        sd.apply( left, right );
        vector<double> s = diag_elements( A );
        reject_elements( s, A(0,0)*MINRELW );
        vector<double> x = element_prod( subrange( prod( left, b ), 0, s.size() ), apply_to_all( s, functor::inv<double>() ));
        row( m_dcf, ipow ) = prod( right, x );
    }
    m_spectrum = spectral_type;
}

integrals::result_type integrals::atmospheric( const matrix<double>& inds, const matrix<double>& desi, double secz ) {
    matrix<double> aints = prod( m_dcf, trans(inds) );           // transformation of the scinds to atmospheric moments
    unsigned int nint = aints.size2();
    vector<double> free_seeing( nint );
    vector<double> isoplanatic( nint );
    vector<double> altitude_eff( nint );
    vector<double> time_const( nint );
    std::transform( row(aints,0).begin(), row(aints,0).end(), free_seeing.begin(), to_seeing( secz, m_lambda_eff ) );
    std::transform( row(aints,1).begin(), row(aints,1).end(), isoplanatic.begin(), to_isopla( secz, m_lambda_eff ) );
    std::transform( row(aints,1).begin(), row(aints,1).end(), row(aints,2).begin(), altitude_eff.begin(), to_altitude( secz ) );
    std::transform( column(desi,0).begin(), column(desi,0).end(), time_const.begin(), to_timeconst( secz ) );
                    // Data filtering is ommited now
                    // Averageing and errors computation
    return result_type(
	result( free_seeing ),
	result( isoplanatic ),
	result( altitude_eff ),
	result( time_const ),
	result( row(aints,0) ),
	result( row(aints,3) )
	);
}
integrals::atmospheric_struct::atmospheric_struct( datum fseeing, datum isangle, datum altieff, datum ticonst, datum moment0, datum moment2 ):
	 fseeing(fseeing), isangle(isangle), altieff(altieff), ticonst(ticonst), moment0(moment0), moment2(moment2) {
}
