/*
    $Id: main.cpp 53 2009-11-01 18:03:28Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <algorithm>
#include <iostream>
#include <map>
#include <memory>
#include <string>

#include <boost/algorithm/string.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/program_options.hpp>

#include "atmos.h"
#include "indices.h"
#include "integrals.h"
#include "params.h"
#include "file_input.h"
#include "file_output.h"
#include "spectral_factory.h"
#include "profile.h"
#include "weif.h"
#include "vector_generator.h"

using namespace boost::numeric;
namespace po = boost::program_options;

namespace {
	template<class T> class nocase_less {
	public:
		typedef T string_type;
	private:
		static bool nocase_cmp( typename string_type::value_type c1, typename string_type::value_type c2 ) {
			return std::toupper( c1 ) < std::toupper( c2 );
		}
	public:
		nocase_less() {}
		/* it is a less function */
		bool operator()( const string_type& s1, const string_type& s2 ) const {
			return std::lexicographical_compare( s1.begin(), s1.end(), s2.begin(), s2.end(), &nocase_cmp );
		}
	};
	class input: public file_input {
	private:
		class measurement_accumulator {
			private:
				input& m_parent;
			public:
				measurement_accumulator( input& parent ): m_parent(parent) {};
				virtual ~measurement_accumulator() {};
			//
				virtual void accumulate( const Date& date, const Time& time, const ublas::vector<double>& mean, const ublas::vector<double>& vars, const ublas::vector<double>& rho1, const ublas::vector<double>& rho2 ) = 0;
				virtual void flush() = 0;
			protected:
				input& operator()() { return m_parent; }
		};
		class dummy_accumulator: public measurement_accumulator {
			public:
				dummy_accumulator( input& parent ): measurement_accumulator(parent) {};
				~dummy_accumulator(){};

				void accumulate( const Date& date, const Time& time, const ublas::vector<double>& mean, const ublas::vector<double>& vars, const ublas::vector<double>& rho1, const ublas::vector<double>& rho2 ) {}
				void flush() {}
		};
		class background_accumulator: public measurement_accumulator {
			public:
				typedef ublas::vector<double>             vector_type;
				typedef ublas::vector<double>::size_type  size_type;
				typedef ublas::vector<double>::value_type value_type;
			private:
				vector_type m_mean;
				size_type   m_num;

			public:
				background_accumulator( input& parent ):
					measurement_accumulator(parent),
					m_num(0), m_mean( atmos::channels ) {}
				~background_accumulator() {}

				void accumulate( const Date& date, const Time& time, const ublas::vector<double>& mean, const ublas::vector<double>& vars, const ublas::vector<double>& rho1, const ublas::vector<double>& rho2 ) {
					if( m_num == 0 )
						m_mean = mean;
					else
						m_mean = m_mean * ( value_type(m_num) / value_type(m_num+1) ) + mean * ( value_type(1)/(m_num+1) );
					m_num++;
				}
				void flush() {
					(*this)().m_indices.set_background( m_mean );
					m_num = 0;
				}
		};
		class normal_accumulator: public measurement_accumulator {
			public:
				typedef ublas::vector<double>             vector_type;
				typedef ublas::vector<double>::size_type  size_type;
				typedef ublas::vector<double>::value_type value_type;
			private:
				Date m_date;
				Time m_time;
				double m_mean_mz;
				double m_mjdfirst;
				double m_mjdlast;
				size_type m_num;

				void dorun( const Date& date, const Time& time, double mz, double moment ){
					Date mdate( floor(moment) );
					Time mtime( 24*(moment-std::floor(moment)) );

					try {
						(*this)().m_outp.normal( date, time, mz );
						// I, F
						(*this)().m_outp.indices( mdate, mtime, (*this)().m_indices.result_indices(), (*this)().m_indices.result_indesis() );
						(*this)().m_outp.fluxes( mdate, mtime, (*this)().m_indices.result_fluxes() );
						// A
						(*this)().m_outp.atmosphere( mdate, mtime, (*this)().m_integrals.atmospheric( (*this)().m_indices.index(), (*this)().m_indices.indesis(), mz ) );
						restoration::result_type profile = (*this)().m_restoration.calcn2((*this)().m_indices.index(), mz, (*this)().m_spectral_factory.spectral((*this)().m_sclass,(*this)().m_pproxy.inner_diam(),(*this)().m_pproxy.outer_diam()) );
						// T, R
						(*this)().m_outp.profile( mdate, mtime, profile.cn2method, profile.cn2seeing, profile.chi2, profile.cn2, profile.ecn2, (*this)().m_restoration.zcn2() );
						(*this)().m_outp.residual( mdate, mtime, profile.cn2method, profile.o_csi );
					} catch( ... ) {
						(*this)().m_indices.reset(); // we have to reset our indices
						throw;
					}
					(*this)().m_indices.reset();
				}
			public:
				normal_accumulator( input& parent ):
					measurement_accumulator(parent),
					m_mean_mz(0), m_mjdfirst(0), m_mjdlast(0), m_num(0) {
				}
				~normal_accumulator() {}
				void settime( const Date& date, const Time& time ){
					m_date = date;
					m_time = time;
				}

				void accumulate( const Date& date, const Time& time, const ublas::vector<double>& mean, const ublas::vector<double>& vars, const ublas::vector<double>& rho1, const ublas::vector<double>& rho2 ) {
					if( m_mjdfirst == 0 )
						m_mjdfirst = (int)date + time/24.0;
					m_mjdlast = (int)date + time/24.0;

					m_num++;
					m_mean_mz = m_mean_mz * ( value_type(m_num-1) / value_type(m_num) ) + (*this)().m_star.Mz( date, time ) * ( value_type(1)/m_num );
					(*this)().m_indices.calculation( mean, vars, rho1, rho2 ); // FIXME:
				}
				void flush() {
					dorun( m_date, m_time, m_mean_mz, 0.5*(m_mjdfirst + m_mjdlast) );// FIXME:

					m_mjdfirst = m_mjdlast = 0;
					m_mean_mz = 0;
					m_num = 0;
				}
		};
	private:
	friend class normal_accumulator;
	friend class background_accumulator;
		typedef parameters_container< std::map< std::string, std::string, nocase_less< std::string > > >  parameters_container_type;
		typedef parameters_proxy< parameters_container_type >                 parameters_proxy_type;

		parameters_container_type m_pcontainer;
		parameters_proxy_type     m_pproxy;

		spectral_factory m_spectral_factory;
		file_output&     m_outp;
		indices          m_indices;
		integrals        m_integrals;
		restoration      m_restoration;
		Coord<equ>       m_star;
		std::string      m_sclass;

		std::auto_ptr< measurement_accumulator > m_maccum;

		template<class T> void switch_accumulator() {
			std::auto_ptr< measurement_accumulator > new_accumulator(new T(*this));
			std::swap( m_maccum, new_accumulator );
			new_accumulator->flush();
		}
	public:
		//FIXME: sending link to not yet constructed object m_maccum(new dummy_accumulator(*this))
		template<class G> input( file_output& out, const spectral_factory& sf, const G& grid):
			m_outp(out),
			m_spectral_factory(sf),
			m_pproxy(m_pcontainer),
			m_maccum(new dummy_accumulator(*this)),
			m_restoration( grid ) {
		}
		~input() {}

		void flush_accumulator() {
			m_maccum->flush();
		}
		void on_comment( const std::string& line ) {
			m_outp.comment( line );
		}
		void on_parameter( const std::string& key, const std::string& value ) {
			m_pcontainer[ key ] = value;

			if( m_pcontainer.key_equal(key,"operations/normal mode/exposition") ){
				m_indices.set_microexposure( m_pproxy.exposition() );
			}
		}
		void on_object( const file_input::object& obj ) {
			m_star = Coord<equ>( obj.coordinates.ra(), obj.coordinates.dec(), m_pproxy.site() );
			m_sclass = obj.sclass;
			m_integrals.coefficients(obj.sclass,m_spectral_factory.spectral(obj.sclass,m_pproxy.inner_diam(),m_pproxy.outer_diam()));

			m_indices.set_counters( m_pproxy.nonlinearity(), m_pproxy.nonpoisson(), m_pproxy.scatter() );
		}
		void on_background( const Date& date, const Time& time ) {
			switch_accumulator<background_accumulator>();
		}
		void on_normal( const Date& date, const Time& time ) {
			switch_accumulator<normal_accumulator>();
			dynamic_cast<normal_accumulator*>(m_maccum.get())->settime(date,time);
		}
		void on_flux( const Date& date, const Time& time ) {
			switch_accumulator<dummy_accumulator>();
		}
		void on_measurements( const Date& date, const Time& time, const ublas::vector<double>& mean, const ublas::vector<double>& vars, const ublas::vector<double>& rho1, const ublas::vector<double>& rho2 ) {
			m_maccum->accumulate( date, time, mean, vars, rho1, rho2 );
		}
	};

	class output: public file_output {
	public:
		output( std::ostream& stm ): file_output( stm ) {}
		~output() {}

		void fingerprint() {
			boost::io::ios_all_saver asaver( (*this)() );

			(*this)() << (boost::format("# %1% %2%") % PACKAGE % VERSION) << std::endl;
		}
	};

	struct profile_grid {
		ublas::vector<double> vec;

		profile_grid( const ublas::vector<double>& vec ): vec(vec) {}
	};
	void validate(boost::any& v, const std::vector<std::string>& values, profile_grid* target_type, int) {
		typedef std::vector< std::string > split_vector_type;

		std::string s = po::validators::get_single_string(values);
		split_vector_type split_vec;

		try{
		boost::trim_if( s, boost::is_space() );
		boost::split( split_vec, s, boost::is_any_of(", "), boost::token_compress_on );
		if( split_vec.size() == 4 ){
			double start = boost::lexical_cast<double>( split_vec[0] );
			double stop  = boost::lexical_cast<double>( split_vec[1] );
			double step  = boost::lexical_cast<double>( split_vec[2] );
			double step2 = boost::lexical_cast<double>( split_vec[3] );
			if( step2 == 0.0 ){
				wf::linear_generator gen( start, step );
				v = boost::any( profile_grid(vector_generator< wf::linear_generator > ( gen, gen.size( stop ) )) );
			} else {
				wf::lg_generator gen( start, step2, step );
				v = boost::any( profile_grid(vector_generator< wf::lg_generator > ( gen, gen.size( stop ) )) );
			}
		} else {
			ublas::vector<double> vec( split_vec.size() );
			for( split_vector_type::size_type it = 0; it < split_vec.size(); ++it ) {
				vec(it) = boost::lexical_cast<double>( split_vec[it] );
			}
			v = boost::any( profile_grid(vec) );
		}
		} catch( const std::exception& e ) {
			throw po::validation_error("invalid value");
		}
	}
};

int main( int argc, char **argv ) {
	double force_accumtime;
	std::string spectra_dir;
	std::string response_file;

	po::variables_map vm;
	po::options_description desc("Allowed options");
	desc.add_options()
		( "help,h",    "produce help message" )
		( "version,v", "produce version message" )
		( "accum-time,a",   po::value<double>(&force_accumtime)->default_value(0), "force accumulation time" )
		( "profile-grid,Z", po::value<profile_grid>()->default_value( profile_grid( vector_generator< wf::lg_generator >( wf::lg_generator( 0.5, 1.0, 2.0 ),6 ) ), "0.5,1,2,4,8,16" ), "set for profile grid generator" )
		( "response,r",     po::value<std::string>(&response_file)->default_value("mass_maid.crv"), "file with detector response" )
		( "spectra-dir,s",  po::value<std::string>(&spectra_dir)->default_value("spectra"), "spectra-files directory" );

	try {
		po::store( po::parse_command_line( argc, argv, desc), vm );
		po::notify( vm );
		if( vm.count("version") ) {
			std::cerr << (boost::format("%1% %2%") % PACKAGE % VERSION) << std::endl;
			std::cerr << "Copyright (C) 2009  Sternberg Astronomical Institute, MSU" << std::endl;
			std::cerr << "This is free software; see the source for copying conditions.  There is NO" << std::endl;
			std::cerr << "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE." << std::endl;
			return 1;
		}
		if( vm.count("help") ) {
			std::cerr << desc << std::endl;
			return 1;
		}

		output outp(std::cout);
		input inp( outp, spectral_factory(boost::filesystem::path(spectra_dir),boost::filesystem::path(response_file)), vm["profile-grid"].as< profile_grid >().vec );
		outp.fingerprint();
		inp(std::cin);
		inp.flush_accumulator();
	} catch ( const po::error& e ){
		std::cerr << e.what() << std::endl;
		std::cerr << desc << std::endl;
		return 1;
	} catch ( const file_input::failure& e ) {
		std::cerr << e.what() << " at input stream, line " << e.line() << std::endl;
		return 1;
	} catch ( const std::exception& e ) {
		std::cerr << e.what() << std::endl;
		return 1;
	}
	return 0;
}
