/*
    $Id: nr.h 53 2009-11-01 18:03:28Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*--------------------------------------------------------------------------*/
/** @file nr.h
                  MASS project: TURBINA module file
                       header file for nr.c

 The module NR contains the set of procedures from "Numerical Recipes in C: The
 art of scientific programming". All the files of the used recipes are
 concatenated in nr.c by means of #include directives. At the top of this concatenation, in this header file,
 we set a heading
 \verbatim #define double double \endverbatim
 to make all the mathematics done in double-precision, as accepted in MASS
 Software (see nrutil.h documentation).

 The memory- and error-handling utilities are given in a separate module
 NRUTIL.

 \b Usage:

 The memory allocation within the MASS "scientific" modules (scind, weif, scan
 and atmos) is done with dvector(), dmatrix() for doubleing point arrays;
 deallocation is made with respective free_dvector() and free_dmatrix(). Thus,
 we use explicitly the double-type arrays.

 The only modified scientific recipe is the \em qromo() integrator which is
 given in qromo2.c source file to implement the relative convergence criterion.
 The use of all the other recipes is normal, as described in the Book. They are
 only modified to conform C++ conventions and return after nrerror() call.

 @version 1.5: Adaptation of recipes to be compatible with C++              */
/* This header file is equipped with documentation made in comments in a
* format suitable for production of LaTeX or HTML documentation with the
* doxygen software.
* Editor tabular size: 4                                                    */
/****************************************************************************/


#ifndef _NR_H
#define _NR_H

/** All the mathematics with "double"-type is done actually in double-type in
nr.c, nrutil.cpp and MASS applications */

#include <boost/numeric/ublas/vector.hpp>

namespace ublas = boost::numeric::ublas;

namespace nr {
    void  four1( ublas::vector<double>& );
    void  hunt(double *xx, int n, double x, int *jlo);
    double midpnt(double (*func)(double), double a, double b, int n);
    void  polint(double *xa, double *ya, int n, double x, double *y, double *dy);
/**
	@brief	Modified integrator for weight calculations
	@param	...	see the NR Book
	@param	EPS	>0: absolute convergence, as in original qromo()
	where EPS is a definition;
	<0: signals to use the relative convergence
	@return void
*/
    double qromo(double (*func)(double), double a, double b, double (*choose)(double (*)(double),double,double,int), double EPS);
                                            // Last "double" stays for EPS in modified proc.in qromo2.c
    void  spline( const ublas::vector<double>& x, const ublas::vector<double>& y, double yp1, double ypn, ublas::vector<double>& y2 );
    void  splint( const ublas::vector<double>& xa, const ublas::vector<double>& ya, ublas::vector<double>& y2a, double x, double& y );
}

#endif // _NR_H
