/*
    $Id: profile.cpp 53 2009-11-01 18:03:28Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <lsp/nnls.h>
#include <lsp/cholesky_decomposition.h>
#include <boost/format.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/triangular.hpp>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <numeric>

#include "io.h"
#include "profile.h"
#include "typesdef.h"

const double MAXCHI2 = 999.0;
const double JPOW_SEE = 0.6 ;
const double LPOW_SEE = -0.2;
const double KS = 1.73588e+07 ;

inline double seeing(double cn2, double secz, double leff ) {
    return KS*pow( cn2/secz, JPOW_SEE )*pow( leff, LPOW_SEE );
}

using namespace boost::numeric::ublas;

template<class T> int zero_elements( const vector<T>&x ) {
    int n = 0;
    for( unsigned int j = 0; j < x.size(); j++ ) n += ( x(j) == 0 )? 1 : 0;
    return n;
}

template<class T> vector<T> element_invert( const vector<T>&x ) {
    vector<T> y = scalar_vector<T>(x.size(),1);
    for( unsigned int j = 0; j < x.size(); j++ ) y(j) /= x(j);
    return y;
}

template<class T> vector<T> diag_elements( const symmetric_matrix<T, upper >&x ) {
    typename matrix<T>::size_type r = x.size1();
    slice s(0, 1, r );
    return matrix_vector_slice< const symmetric_matrix<T, upper > > ( x, s, s );
}

restoration::~restoration(){
}

void restoration::covariance( const matrix<double>& in,  vector<double>& mean, symmetric_matrix< double, upper >& cov ) const {
    const double damper = 0.8;
    unsigned int n = in.size1();
    unsigned int k = in.size2();
    for( unsigned int i = 0; i < k; i++ ) mean(i) = sum(column( in, i ))/(double)n;
    for( unsigned int i = 0; i < k; i++) for( unsigned int j = i; j < k; j++)
        cov(i,j) = inner_prod( column( in, i ), column( in, j ))/n - mean(i)*mean(j);
    for( unsigned int i = 0; i < k; i++) for( unsigned int j = i+1; j < k; j++)
        cov(i,j) = damper*cov(i,j);
    if( zero_elements( diag_elements( cov ) ) > 0 ) {
	throw failure( (boost::format("Zero variances: %1% x %2% %3% %4%") % n % k % mean % vector<double>(diag_elements( cov )) ).str().c_str() );
    }
}

double restoration::nnls_solve( const matrix< double >& Wf, const matrix<double>& inds,
       vector<double>& cn2, vector<double>& err, vector<double>& rsd ) const {
    double ret = 0;
    unsigned int nw  =  Wf.size1();
    unsigned int ncn2 = Wf.size2();
    unsigned int nidx = inds.size1();
    assert( nw > 1 );
    assert( ncn2 > 1 );
    matrix< double > weights = identity_matrix< double >( nw );

        if( nidx < 2 ) {
	    throw failure( (boost::format("Too few measurements: %1% x %2%") % nidx % inds.size2() ).str() );
        }
        vector<double> mean = zero_vector< double >( nw );
        symmetric_matrix< double, upper > covars( nw, nw );
        covariance( inds, mean, covars );
        lsp::cholesky_decomposition( covars, upper_tag() );             // covars  is broken now
        triangular_adaptor< symmetric_matrix< double, upper >, lower > U(covars);
        inplace_solve( U, weights, lower_tag());
        for( unsigned int i = 0; i < nw; i++ )
            if( weights(i,i) == 0 || isinf(weights(i,i)) || isnan(weights(i,i)) ) {
		  throw failure( (boost::format("Zero or infinity weights: %1%") % vector<double>(diag_elements(weights)) ).str() );
            }
        matrix< double > A = prod( weights, Wf );
        matrix< double > profile( nidx, ncn2 );
        vector< double > sc2( nidx );
        vector< double > e2 = zero_vector< double >( ncn2 );
        vector< double > r2 = zero_vector< double >( nw );
        cn2 = zero_vector< double >( ncn2 );
        for( unsigned int k = 0; k < nidx; k++ ) {
            vector< double > b = prod( weights, row( inds, k) );
            matrix< double > cov = zero_matrix<double>(ncn2, ncn2);
            lsp::nnls< matrix< double >, vector< double >  > nnls( A, b );
            vector< double > y;
            nnls.solve( y , cov );
            vector< double > r = prod( A, y ) - b;
            sc2( k ) = inner_prod( r, r );
            r2 += r;
            cn2 += y/(double)nidx;
            e2 += element_prod( y, y );
            row( profile, k ) = y;
        }
        err = ( e2/(double)nidx - element_prod( cn2, cn2 ))/(double)(nidx-1);
        rsd = element_prod( prod( U, r2/(double)nidx), apply_to_all( mean, functor::inv<double>() ));
        ret = sum( sc2 )/(double)nidx;

    return ret;
}

restoration::result_type restoration::calcn2( const matrix<double>& ind, double secz, const Weights& basewf  ) const {
    Weights wfs( basewf.apertures_number(), m_userzcn2*secz );
    wfs.interpolate( basewf ) ;
    profile_struct ret;
                                    // Solve the task by nnls method
    if( ( ret.chi2 = restoration::nnls_solve( wfs.weights(), ind, ret.cn2, ret.ecn2, ret.o_csi ) ) < 0 ) {
        ret.cn2.clear();
        ret.ecn2.clear();
    } else {
        ret.cn2 /= secz;
        ret.ecn2 /= secz;
    }
    ret.chi2 = std::min( ret.chi2, MAXCHI2 );
                                    // compute and return the seeing which corresponds to restored profile.
                                    // Zenith distance is set to 1 since Cn2[] is already corrected for it
    ret.cn2seeing = seeing( sum( ret.cn2 ), 1, basewf.lambda_eff() );
    ret.cn2method = "X";
    return ret;
}
