/*
    $Id: weif.cpp 56 2009-11-01 18:19:52Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/****************************************************************************/
/** @file weif.cpp
                 MASS project: TURBINA module file
                             weif.cpp
                    Description is given in weif.hxx                        */
/* Editor tabular size: 4                                                   */
/****************************************************************************/

/* Insert the definition WEIFTEST in compilation for the module self-tests:
gcc -c weif.c .... -DWEIFTEST=1 */

#include <fstream>
#include <iostream>
#include <iomanip>
#include <vector>
#include <algorithm>
#include <functional>
#include <numeric>
#include <cmath>
#include <limits>

#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>

#include "io.h"
#include "weif.h"
#include "nr.h"

#ifdef WEIFTEST

/*------------------------------ TESTING PART --------------------------*/
/* We test the module with the main() part included: see definition WEIFTEST */
/*----------- Comes normally from configuration file: cf_... vars ----------*/
/** Usual value for low boundary of altitude range */
const double Z0 = 0 ;

/** Maximal feasible value for upper boundary of altitude range. Since
"altitude" in weight functions is a length along line-of-sight in atmospheric
calculations, this ZMAX=40 [km] stays for an altitude of 20 km at airmass=2. */
const double ZMAX = 40 ;

/** Optimal altitude modifier for z-grid creation in proportional mode */
const double DZ = 0.2 ;

/** Optimal minimal altitude step for z-grid creation in proportional mode */
const double DZMIN  = 0.04 ;

const int max_nodes = 20;               // number of apertures in segmentor :
const int max_naper = 10;               // number of apertures in segmentor :
                            // Effective size of segmentor annular apertures [cm] :
double cf_aper_inn[max_naper] = { 0.0000,1.6485,3.0615,5.7304999};
double cf_aper_out[max_naper] = { 1.6171,2.9830,5.6520,10.5975};
                                            // Magnification of the system (multiplier of cf_aper_* sizes):
double cf_magnif = 1.0;
                                            // Altitude grid:
double cf_z0 = Z0, cf_zmax = ZMAX, cf_dz = DZ, cf_dzmin = DZMIN;
                                            // Alternative grid: to interpolate to
double int_z0 = 0, int_zmax = 20.0, int_dz = 0.02, int_dzmin = 0;

template<class T, int nmax > int str2da( T* a, const std::string& cs ) {
    std::string s(cs);
    std::string::size_type i = 0;
    while( (i = s.find_first_of( ",", i )) != std::string::npos ) s.at(i) = ' ';
    int n = 0;
    std::istringstream istr(s);
    while( n < nmax && istr >> a[n] ) n++;
    return n;
}

int main( int argc, char **argv ) {
    int cf_naper = 4;
    int nodes = 0;

    std::string rfile = "/opt/turbina/data/data/mass.crv";
    std::string sfile;
    std::string wfile = "/dev/stdout";
    double d[max_naper];
    double zgrid[max_nodes];
	int c;                                  // parse options:
	while( ( c = getopt(argc,argv,"r:s:N:Z:I:O:hv") ) != -1  ) {
		if( argc < 3 ) c = 'h';             // no spectra and sed file(s) - output help
		int nap = max_naper;
		switch (c) {
			case 'r':    rfile = std::string(optarg); break;
			case 's':    sfile = std::string(optarg); break;
			case 'N':  { int n=atoi(optarg); if ( n>0 && n<=max_naper ) cf_naper = n; break; }
			case 'I':    nap = str2da<double,max_naper>( cf_aper_inn, std::string(optarg) ); break;
			case 'O':    nap = str2da<double,max_naper>( cf_aper_out, std::string(optarg) ); break;
			case 'Z':    nodes = str2da<double,max_nodes>( zgrid, std::string(optarg) );  break;
			case 'h':
            default :
				std::cerr << "WEIF: scintillation weight functions computer" << std::endl;
				"Usage: \n./weif -ssfile [-rrfile] [-Nn] [-Ii1,..in -Oo1,..on] [-Zz1,z2,z3,z4] >out.wf\n"
				"  Input data files:\n"
				"\t-s spectrum-file \n"
				"\t-r response-file (default: /opt/turbina/data/data/mass.crv)\n"
				"  Options:\n"
				"\t-N number of apertures (default: 4)\n"
				"\t-I innerDiam[0],...,innerDiam[N-1] (in cm, default: old MASS @k=15.7)\n"
				"\t-O outerDiam[0],...,outerDiam[N-1] (in cm, default: old MASS @k=15.7)\n"
				"\t-Z zmin,zmax,zstep,zminstep - altitude grid parameters, km \n\t\t(default: 0,40,0.2,0.04)\n";
				exit (-1);
		}
	}
    Weights cf_wf( cf_naper, wf::grid_generate( nodes, zgrid ) );
    vector<double> cf_eps_inner(cf_naper);
    vector<double> cf_eps_outer(cf_naper);
    for( int i = 0; i < cf_naper; i++ ) cf_eps_inner(i) = cf_aper_inn[i]*cf_magnif;
    for( int i = 0; i < cf_naper; i++ ) cf_eps_outer(i) = cf_aper_out[i]*cf_magnif;
    Spectral_data spec( rfile, wfile, sfile );
	cf_wf.calculation( spec, cf_eps_inner, cf_eps_outer );
    return 0;
}

#endif                    // ifdef WEIFTEST end
/*----------------------- TESTING PART END --------------------------------*/
                            // local functions for pure elementary calculations of weights:
namespace wf {
    void make_fft( vector<double>&, vector<double>& ) ;
    double sfunc( double );
    double sdum( double );
    double flama( double ) ;
    double weight( double z, double eps1, double eps2, double eps3, double eps4) ;
    void getspechar( vector<double>& lambda, vector<double>& edist, double& leff, double& lblue, double& lred );
    const double DER0  = 0.0 ;          // Splines: define the first derivative of weights at z=0 equal to 0:
    const double DERNAT  = 1.0E31 ;     // signal to set the "natural" upper boundary cond.: 2nd weight der.=0
    const double MKM2CM	 = 1E-4 ;       // Conversion of [mkm] to [cm]:
}

/***************************************************************************/
/*                  Public functions implementation                        */
/***************************************************************************/

/*------------------------- Weights::constructor -----------------------------------*/

Weights::Weights( int n, const vector<double>& z ) : mass_nchan(n), weights_number(n*(n+1)/2), zgrid(z) {
    ws.resize( weights_number, zgrid.size() );
}

void Weights::write( const std::string& filename ) const {
    static std::string name[] = { "A", "B", "C", "D", "AB", "AC", "AD", "BC", "BD", "CD" };
    assert( ws.size1() != 0 && ws.size2() != 0 );
    std::fstream file;
    file.open( filename.c_str(), std::ios::out );
    if( !file ) throw std::runtime_error( "File " + filename + " open error" );
    if( !(file << "# N " << eps_inn.size() << std::endl) ) throw std::runtime_error( "File " + filename + " write error" );;
    for( int i = 0; i < eps_inn.size(); i++ ) {
        file << "# " << name[i] << " " << std::fixed << std::setprecision(2) << eps_inn(i) << " " << eps_out(i) << std::endl;
    }
    file << "# S " << std::fixed << std::setprecision(4) << leff << " " << lblue << " " << lred << std::endl;
    file << "# Z " << std::setprecision(2) << zgrid << std::endl;
    file << "# W " << std::setprecision(0) << std::scientific << wf::WSCALE << std::endl;
    file << "##Altit.";
    for( int i = 0; i < ws.size1(); i++ ) file << " W(" << name[i] << ")/" <<  std::scientific << wf::WSCALE;
    file << std::endl;
	for( int i = 0; i < zgrid.size(); i++) {
        file << std::fixed << std::setprecision(3) <<  std::setw(7) <<  zgrid(i) << " ";
		for( int j = 0; j < ws.size1(); j++) {
            file << " " << std::scientific << std::setprecision(4) <<  std::setw(10) << ws(j,i)/wf::WSCALE;
		}
        file << std::endl;
    }
}

template<class T>  bool more_dense( const vector<T>& x, const vector<T>& y ) {
    if( x.size() == 0 ) return false;
    if( y.size() == 0 ) return true;
    return x(0) < y(0) && x( x.size()-1 ) > y( y.size()-1 ) && x.size() > y.size();
}

void Weights::interpolate( const Weights& refwf ) {
           // Check entries and parameters: new_wf zgrid should be enclosed in z-range of the old wf grid!
    assert( refwf.ws.size1() == refwf.weights_number && refwf.ws.size2() == refwf.zgrid.size() );
    assert( zgrid.size() > 0 );
    if( more_dense( zgrid, refwf.zgrid ) ) throw std::runtime_error( "Attempt to interpolate from narrower zgrid" );
    ws.resize( refwf.ws.size1(), zgrid.size() );
                                                            // Spline interpolation :
    vector<double> y2( refwf.zgrid.size() );
    for (int w = 0; w < refwf.ws.size1(); w++ ) {           // Zero value of weight (not shifted) => zero (theoretic) deriv.:
        double start_deriv = ( refwf.ws(w,0) != 0 ) ? wf::DERNAT : wf::DER0;
        vector<double> rw = row( refwf.ws, w );
        nr::spline( refwf.zgrid, rw, start_deriv, wf::DERNAT, y2 );
        for (int iz = 0; iz < zgrid.size(); iz++) nr::splint( refwf.zgrid, rw, y2, zgrid(iz), ws(w,iz) );
    }
                                                            // Interpolation is successful: assign fields:
    leff = refwf.leff;
    lblue = refwf.lblue;
    lred = refwf.lred;
    eps_inn = refwf.eps_inn;
    eps_out = refwf.eps_out;
}

/**
	@brief	Compute some characteristic points on the spectrum
	@param	lambda 	wavelength grid
	@param	edist	respective spectral energy distribution
	@param	nedist	length of \em lambda and \em edist grids
	@param	leff	Returned effective wavelength in [units of \em lambda]
	@param	lblue	Returned wavelength of the 50\% level of SED at blue side
			from the SED's peak in [units of \em lambda]
	@param	lred	ditto at the red side from the SED's peak

	Function searchs for the maximum of \em edist[] computing simultaneously
	the gravity center of \em edist to be returned as \em leff. Then, it goes
	to the left and to the right from the maximum until the level of 50\% level
	of the maximal one is crossed by the curve. The \em lblue and \em lred
	points are computed by the local linear interpolation between the adjacent
	points.

	@see wf_leff()
*/

void wf::getspechar( vector<double>& lambda, vector<double>& edist, double& leff, double& lblue, double& lred ) {
    assert( edist.size() == lambda.size() );
    leff = inner_prod( lambda, edist )/ sum( edist );
    if( isinf(leff) || isnan(leff) ) throw std::runtime_error( "Illegal response data" );
    vector<double>::iterator max_pos = max_element( edist.begin(), edist.end() );
    double emax = *max_pos;
    int l =  max_pos.index();
	while( l-- > 0 &&  edist(l) > emax/2 );
    lblue = lambda(l) + ( emax/2 - edist(l) )*( lambda(l+1) - lambda(l) )/( edist(l+1) - edist[l] );
    l = max_pos.index();
    while( ++l < lambda.size() &&  edist(l) > emax/2 );
    lred = lambda(l-1) + ( edist(l-1)-emax/2 )*( lambda(l) - lambda(l-1) )/( edist(l-1) - edist(l) ) ;
}

int Weights::calcleff( vector<double>& lambda, vector<double>& edist ) {
    wf::getspechar( lambda, edist, leff, lblue, lred ) ;
}

void Weights::calc( vector<double>& lambda, vector<double>& edist ) {
    double dlambda[100], dedist[100];
    int nedist = edist.size();
    assert( zgrid.size() >  0 );
    calcleff( lambda, edist );                      // Call the SED's effective wavelength and 50% points computer at last
    wf::make_fft( lambda, edist ) ;
    for( int i = 0; i < zgrid.size(); i++ ) {
        int nnw = 0;                                // Counter of weighting functions
        for (int a = 0; a < mass_nchan; a++) {      // Normal weighting functions:
            ws(nnw,i) = wf::weight( zgrid(i), eps_out(a), eps_inn(a), 0, 0 );
            nnw++;
        }
        for (int a = 0; a < mass_nchan - 1; a++) {  // Differential weighting functions:
            for (int b = a + 1; b < mass_nchan; b++) {
                ws(nnw,i) = wf::weight( zgrid(i), eps_out(a), eps_inn(a), eps_out(b), eps_inn(b) );
                nnw++;
            }
        }
    }
}

/*==================== weight calculation section =====================*/

namespace wf {
                        // Global private parameters needed to transfer the geometry of apertures into
                        // the function flama() to integrate for weight:
    double diameter1, diameter2, diameter3, diameter4;
    double lambda0 = 0;
                        // Last parameter is the SED origin in [mkm] (wavelength of the first point
                        // in a given SED) which is a shift of SED before doing its Fourier Tranform.
                        // Weight integration: global private arrays
                        // Complex FT of SED, respective frequency grid, its Re and Im parts, their spline coefficients and length:
    vector<double> edist_ft;
    vector<double> edist_freq;
    vector<double> re_edist_ft;
    vector<double> im_edist_ft;
    vector<double> re_edist_ftspln;
    vector<double> im_edist_ftspln;
    int edist_ftnn ;
    double altitude ;	// Current altitude in [cm]
}

/**
	@brief	Compute the FT of SED in units of [1/cm]
	@param	lambda	wavelength grid
	@param	edist	light SED
	@param	nedist	w/l grid length
	@return	0 if Ok, <0 on error

	The Fast Fourrier Transform is used to compute the FT.

	Before FFT, the SED is normalized by division of each value by the integral
	of SED. Then, the grid of SED is expanded four times with padding the obtained
	"tail" with zero values; SED values are divided by their argument
	wavelengths. Thus, the result of the FFT algorithm (four1()
	utility from NR is used) has a better sampling in the frequency domain
	which makes the spline interpolation of the spectrum more robust.

	Since the grid of SED starts at non-zero wavelength but immediately near the
	wavelength of the blue-end of the sensitivity curve of MASS, the FT is made
	on this origin-shifted spectrum. So, the correction for this shift is needed
	afterwards (see below).

	The units of the wavelengths are converted into [cm], the shift of the SED
	to its origin and the sampling of the wavelength grid are also expressed in
	[cm]. This ensures the correct normalization of the resulting FT and a
	proper correction of the real and imaginary parts of the FT for the made
	shift of SED to its origin. The latter correction is made in sfunc()
	analytically.

*/

template <class T> T nexthigher(T k) {
    if (k == 0) return 1;
    k--;
    for (int i=1; i < std::numeric_limits<T>::digits; i<<=1) k = k | k >> i;
    return k+1;
}

void wf::make_fft( vector<double>& lambda, vector<double>& edist ) {
                                // Number of stretching twice the F(lambda) grid to improve sampling of its FT
                                // 0 - no stretch, 1 - stretch twice, 2 - stretch 4-times etc.
    const int GRIDFACT = 2 ;
                                // The shift of the SED grid "lambda0" in [cm] (global wf::-var.)
    lambda0 = lambda[0]*MKM2CM;
    vector<double> dlam( lambda.size() );
    std::adjacent_difference( lambda.begin(), lambda.end(), dlam.begin() );
    if( std::count( dlam.begin()+1, dlam.end(), 0 ) > 0 )  throw std::runtime_error( "WEIF: lambda-grid has zero step" );
    double dlambda = dlam(1);   // Sampling of SED:
    double enorm = sum( edist );       // Normalize SED:
    if( enorm <= 0 ) throw std::runtime_error( "WEIF: SED has non-positive norm" );
    edist /= enorm;
                                // Make the least number which is a power of two and not less than nedist:
                                // The length "nn" of the data grid for FFT is meanwhile 2**GRIDFACT
                                // times more than this least number.
//    edist_ftnn = 1 << GRIDFACT;
//    for( int l = edist.size(); l != 0; edist_ftnn <<= 1, l >>= 1 ){ }
    edist_ftnn = 2*GRIDFACT*nexthigher(edist.size());
//    std::cerr << "fft:ftnn " << edist_ftnn << " edist size " << nexthigher<unsigned int>(edist.size()) << std::endl;
	                           // Make the complex-type SED-function array with the length like 2**N
	edist_ft.resize( edist_ftnn*2 );
	                           // Make the double-type array for frequencies of the FT of SED to be computed:
	edist_freq.resize( edist_ftnn/2 );
    for( int l = 0; l < edist_freq.size(); l++ ) edist_freq(l) = l/(dlambda*MKM2CM)/edist_ftnn;
	                           // Fill in the data array for FFT:
	for( int l = 0; l < edist_ftnn; l++ ) {
		                      // Real part of the complex SED divided by the wavelength [cm]:
		edist_ft[2*l] = ( l < edist.size() ) ? edist[l]/(lambda[l]*MKM2CM) : 0;
		edist_ft[2*l+1] = 0;  // Imaginary part is zero:
	}
                                // Compute the Fourier transform of the origin-shifted SED:
    nr::four1( edist_ft );
    slice re( 0, 2, edist_ftnn/2 );
    slice im( 1, 2, edist_ftnn/2 );
    re_edist_ft = vector_slice< vector<double> >( edist_ft, re );
    im_edist_ft = vector_slice< vector<double> >( edist_ft, im );
                                // Interpolate the positive frequency Re- and Im-parts of FFT(F(lambda)) by splines:
                               // Make double-type arrays for spline coefficients of Re- and Im-parts of the FT of SED:
    re_edist_ftspln.resize( edist_ftnn/2 );
    im_edist_ftspln.resize( edist_ftnn/2 );
	nr::spline( edist_freq, re_edist_ft, wf::DERNAT, wf::DERNAT, re_edist_ftspln );
	nr::spline( edist_freq, im_edist_ft, wf::DERNAT, wf::DERNAT, im_edist_ftspln );
}

/***************************************************************************/
/*                  Private functions implementation                       */
/***************************************************************************/

/**
	@brief	Compute the imaginary part of the Fourier transform of the light SED
	@param	f	frequency in [1/cm] units
	@return	Im(FT(SED))**2

	Private function used by flama().

	The results of the spline approximation of the real and imaginary parts  of
	FT of \f$\lambda_{min}\f$-shifted SED are used. Spline coefficients are
	already computed in make_fft() and put into \em re_edist_ftspln[] and
	\em im_edist_ftspln[] arrays. Re- and Im-parts  are taken from
	splines, corrected for the shift of SED to its origin (made while making
	FFT) by multiplication by \f$ exp(-2\pi i \lambda_{min}) \f$ and the square
	of the  imaginary part of the product is returned.

	Spline coefficients \em re/im_edist_ftspln[], computed FT grid \em
	re/im_edist_ft[] and the frequency grid \em edist_freq[] are the global
	(private) arrays with the length \em edist_ftnn.

	@see Weights::calc()
*/

double wf::sfunc( double f ) {
                       // High frequencies (beyond Nyquist) are cut off:
    if( f > edist_freq[edist_ftnn/2-1] ) return 0 ;
                       // Compute the real and imaginary parts of shifted F(lambda[mkm]):
    double re_ft = 0, im_ft = 0 ;
    nr::splint( edist_freq, re_edist_ft, re_edist_ftspln, f, re_ft );
    nr::splint( edist_freq, im_edist_ft, im_edist_ftspln, f, im_ft );
                       // Correct the FT for shifting, Multiply the numbers taking only the imaginary part
    double tmp = im_ft*cos( 2*M_PI*f*lambda0 ) + re_ft*sin( 2*M_PI*f*lambda0 );
    return tmp*tmp ;
}

/**
	@brief	Test-case of S-function: analytical for "quasi-gaussian" SED
	@param	f	frequency [1/cm]
	@return	S

	Private function which may replace sfunc() in flama() for test-calculations.

	In case of the SED being the "quasi-gaussian" function, the \em S-function
	is a purely analytical expression:
	\f[
	S = \sin^2(2\pi\lambda_0 f) exp(-4(\pi\sigma f^2)^2)/\lambda_0
	\f]
	for the SED which is a normalized \f$\lambda\f$-multiplied gaussian with
	a \f$\sigma\f$-dispersion centered at \f$\lambda_0\f$.

	The substitution of sdum() instead of sfunc() in flama() may help to test
	the influence of \em "non-gaussianity" of the spectrum on the resulting
	shape of the weight function. The effect, in general, does not exceed a few
	percents (below 10 in any case) for the A0V star and the MASS response
	function.

*/

double wf::sdum( double f ) {
                // Put below, directly the FWHM and central wavelength of "quasi-gaussian" in [mkm]:
    const double DELTA = 0.1 ; //0.1
    const double LAMBDA0 = 0.5 ;
	double tmp = sin( 2.0*M_PI*LAMBDA0*MKM2CM*f )/(LAMBDA0*MKM2CM) ;
	tmp *= tmp ;
	tmp *= exp( -7.11941*f*f*(DELTA*MKM2CM)*(DELTA*MKM2CM) );
	return tmp ;
}

/**
	@brief	Compute the under-integral expression for the weight integration
	@param	x	spatial frequency in inverse meters
	@return Under-integral expression (see doc)

	Private function used by weight().

	In polychromatic treatment, the weight is integrated from:
	\f[
	E = f^{-8/3} * S(z f^2 /2) * A(f) .
	\f]

	where the function \em S(f) is a \em squared integral of SED of incident
	light with sine-modulation over wavelengths:

	\f[
	S(f) = \left [ \int\limits_{\lambda_{min}}^{\lambda_{max}}
		\sin(2\pi \lambda f) F(\lambda) d\lambda \right]^2
	\f]
	which is, in fact, the squared imaginary part of the Fourier Transform
	of \f$ F(\lambda)/\lambda \f$ computed by sfunc().

	If SED is a \em delta-function centered at \f$ \lambda_0\f$, then the
	expression turns into the known monochromatic one:

	\f[
	E = f^{-8/3} * \sin^2(z f^2 \lambda_0) * A(f) .
	\f]

	Here A(f) is a normalized aperture OTF for the given geometry of two
	annular apertures. For simplicity, the geometry parameters diameter,
	diameter1..4 are  transferred via the global parameters; function \em
	S() is also the global (private) function sfunc().

*/

const double POWIND = -2.66666666667;

double wf::flama( double x ) {
    double apert = 1.0 ;
    double apert1 = 0.0, apert2 = 0.0, apert3 = 0.0, apert4 = 0;
    double tmp = pow( x, POWIND )*sfunc( altitude*x*x/2 );
//    double tmp = sdum( altitude*x*x/2 );
//    double vtf = 100*x; V=1 m/c,t=1 c
//    if( diameter > 0 ) {
    apert  = x*M_PI;                            // 2pi*f*(r=D/2)
    if( diameter1 != diameter2 ) {                // First aperture:
        apert1 = diameter1*j1( diameter1*apert );
        apert2 = diameter2*j1( diameter2*apert );
        apert1 = ( apert1 - apert2 )/( diameter1*diameter1 - diameter2*diameter2 );
    }
    if( diameter3 != diameter4) {                 // Second aperture: the same with indices 1->3, 2->4:
        apert3 = diameter3*j1( diameter3*apert );
        apert4 = ( diameter4 == 0 )? 0 : diameter4*j1( diameter4*apert );
        apert3 = ( apert3 - apert4 )/( diameter3*diameter3 - diameter4*diameter4 );
    }
    apert = 2.0*( apert1 - apert3 )/apert;
//    }
    return tmp*apert*apert;
}

/**
	@brief	Compute the weight of a given \em z in a scintillation index
	@param	z	altitude [km]
	@param	d	reference size of an aperture [cm]
	@param	eps1	relative outer size of 1st (larger) aperture
	@param	eps2	relative inner size of 1st (larger) aperture
	@param	eps3	relative outer size of 2nd (smaller) aperture (may be 0)
	@param	eps4	relative inner size of 2nd (smaller) aperture (may be 0)
	@return Weight value in [m^(1/3)]

	Private	function used by Weights::calc().

	Function returns the weight for an altitude \em z in the stellar
	scintillation. The FT of the incident light SED divided by \em lambda is
	assumed to be ready and accessible by the function sfunc(). The diameter of
	the largest entrance aperture is \em d and the radii of annular apertures
	are obtained by its multiplication by \em eps1..4.

	The first (say, larger) aperture has the outer diameter \em d*eps1 and the
	inner \em d*eps2. The second (smaller) aperture has the outer diameter \em
	d*eps3 and the inner \em d*eps4.

	If only two of four diameters differ from each other (non-zero ring width)
	then the weighting function value for the  \em normal scintillation index
	is  returned for a given aperture (for the larger one - if eps1>eps2>0,
	eps3=eps4; for the smaller one - if vise versa).

	If three or four diameters are different (eps1>eps2 and eps3>eps4) then the
	weighting finction value for the \em differential scintillation index is
	returned for a pair of apertures with sizes \em (d*eps1,d*eps2) -- \em
	(d*eps3,d*eps4).

	The polychromatic expression for the weight is used here, which uses the FT
	of the light SED to compute the so called \em S-function sfunc(). This
	function stays for the part of the under-integral expression of the weight
	which is computed by flama(). And the latter is already the function
	which is integrated by the qromo() integrator with the \em relative
	convergence criterion. It integrates the weight for the range of spatial
	frequencies from \em X1=1e-4 to \em X2=18.

	The weight for the zero altitude is artificially set to zero.
*/

double wf::weight( double z, double eps1, double eps2, double eps3, double eps4 ) {
                    // Define the range of frequencies to integrate the aperture OTF A(f)
                    // (X1=0.0001, X2=15 correspond to L_0=30m and l_0=5 mm at 10 km at 0.5mkm):
    const double X1 = 0.0001 ;      // 0.0001
    const double X2 = 1.0 ;         // 18  0.8   ( 0.3  11
    const double EPS = -1.E-7 ;     // Relative precision for integration with modified qromo():
/* CALIBR = (32*PI^3*9.69*1E-3)*(0.01^(5/3))*1E4; where 0.01^(5/3) comes from
integration in [cm] instead of [m]; 1E4 originates from conversion of [1/cm]^2
into [1/m]^2 in integration of sfunc() */
    const double CALIBR  = 44.62621363 ;
    const double KM2CM  = 1E5 ;     // Conversion of [km] into [cm]:
                                    // Assign the global parameters:
    wf::altitude = z*KM2CM;         // Altitude in [cm]
    wf::diameter1 = eps1;
    wf::diameter2 = eps2;
    wf::diameter3 = eps3;
    wf::diameter4 = eps4;
                                    // CHECK: printout all the aperture functions if "return tmp" is replaced with "return apert" in flama():
                                    // if (z==0) for (w=X1; w<=X2; w*=1.2) {	printf("%8.5f %10.4e\n",w,flama(w)); }	return 0;
    if( std::abs( z ) < std::abs( EPS ) ) return 0;
    else return CALIBR*nr::qromo( flama, X1, X2, nr::midpnt, EPS );
}

namespace wf {
                                // Units of optical range wavelength: 0 for [mkm], 3 for [nm], 4 for [\AA]
    const int MKM_RANGE	= 0	;
                                // File-reader for spectra and response curve
    void read2col( const std::string&, vector<double>&, vector<double>& ) ;
//    void conwgrid( vector<double>&, int );
    void sedcrv( vector<double>&, vector<double>&, vector<double>&, vector<double>& );
}

void Weights::calculation( const Spectral_data& specfile, const vector<double>& ap_inn, const vector<double>& ap_out ) {

    vector<double> rlambda;
    vector<double> response;
    vector<double> slambda;
    vector<double> sed;
    wf::read2col( specfile.response, rlambda, response );   // Read the response function from the file:
    wf::read2col( specfile.spectra, slambda, sed );
    wf::sedcrv( slambda, sed, rlambda, response );          // Check that response curve wavelength grid complies the spectrum and multiply them
    assert( ap_inn.size() == mass_nchan && ap_out.size() == mass_nchan );
    eps_inn = ap_inn;
    eps_out = ap_out;
    calc( rlambda, sed );
   // write( specfile.weights );
}

/**
	@brief	Read the two columns of argument and function values from file
	@param	filename	file name
	@param	col1	pointer to the array of arguments (first column data)
	@param	col2	pointer to the array of function values (second column)
	@param	nrow	pointer to the place where to put the length of allocated
		arrays of arguments and function values

	This utility reads the pairs of numeric values from the text-type file
	ignoring the lines beginning with comment symbol "#". Two non-allocated
	pointers must be supplied as NULLs on start. Function counts the non-comment
	data lines, allocates the memory for arrays of argument (first column) and
	function values (second column) with lengths equal to the number of the
	counted data lines, and then reads the columns into them.

	The following checks are done on each non-comment line:
	- at least two numeric tokens are given in each line (argument and function)
	- argument is a monotonously rising function
	- argument values start from some positive number
	- function is non-negative
	- function has a positive average
	The function is thus suited for reading the spectrum-like information.

	Errors possible are the problem of file opening/reading or line parsing. In
	case of error, the pointers are returned unallocated.
*/

void wf::read2col( const std::string& filename, vector<double>& col1, vector<double>& col2  ) {
    std::fstream file;
    file.open( filename.c_str(), std::ios::in );
    if( !file ) throw std::runtime_error( "File " + filename + " open error" );
    std::vector<double> x;
    std::vector<double> y;
    std::string iline;
    while( getline( file, iline ) ) {
        if( iline.size() < 1 || iline.at(0) == '#' ) continue;
        std::istringstream istr( iline );
        double cx, cy;
        if( !(istr >> cx >> cy) )  throw std::runtime_error( "File " + filename + " line format error" );
        x.push_back( cx );
        y.push_back( cy );
    }
    if( x.size() == 0 ) throw std::runtime_error( "File " + filename + " contains no data" );
    if( std::adjacent_find( x.begin(), x.end(), std::greater<double>() ) != x.end() ) throw std::runtime_error( "File " + filename + " contains unsorted data" );
    if( std::accumulate( y.begin(), y.end(), 0.0 ) <= 0.0 ) std::runtime_error( "File " + filename + " contains zero or negative data" );
    col1.resize( x.size() );
    col2.resize( x.size() );
    std::copy( x.begin(), x.end(), col1.begin() );
                            //Convert array of optical (!) wavelengths into given scale
    double conv = pow( 10.0, MKM_RANGE - ((int)log10( col1(col1.size()/2))+1) );
    col1 *= conv;
    std::copy( y.begin(), y.end(), col2.begin() );
}

/**
	@brief	Multiply the spectrum curve by the response curve with
		optionally shifted origins
	@param	slambda	array of wavelengths of spectrum
	@param	sed  	array of spectrum; returned \b multiplied by
		response in first \a nresponse elements
	@param	nsed	length of \a sed[] and \a slambda[]
	@param	rlambda	array of wavelengths of response
	@param	response	array of response curve
	@param	nresponse	length of \a response[] and \a rlambda[]

	Function looks for the coincidence of some \em k-th element of \em
	slambda[] grid with the first element of the response curve wavelength
	grid \em rlambda. This \em k is a shift (or difference) of origins of
	SED and response curves (normally, the spectrum is wider).

	Then, first elements of spectrum - \em sed[i], i=0..nresponse - are
	rewritten with the product of SED and response curves
	\em	sed[i+k]*response[i] which takes into account the different origins of
	these functions. Last \em nsed-nresponse elements of \em sed[] are replaced with
	zeros. Thus, on exit only first \em nresponse elements of \em sed[] make
	sense.

	@see calcwf() checksed()
*/

void wf::sedcrv( vector<double>& slambda, vector<double>& sed, vector<double>& rlambda, vector<double>& response ) {
                            // Find the starting point of the response curve at SED wavelength grid:
    vector<double>::iterator pos;
    pos = std::find( slambda.begin(), slambda.end(), rlambda(0) );
    if( pos ==  slambda.end() || !std::equal( pos, pos+rlambda.size(), rlambda.begin() ) )
                throw std::runtime_error( "No match of response and spectrum grids" );
    int begsed = pos - slambda.begin();
    vector<double> resed = element_prod( subrange( sed, begsed, begsed+response.size() ), response );
    sed = resed;
}

wf::linear_generator::linear_generator( result_type start, result_type step ): start_(start), step_(step) {
}
wf::linear_generator::result_type wf::linear_generator::operator()( size_type i ) const {
	return ( start_ + step_ * i );
}
wf::linear_generator::size_type wf::linear_generator::size( result_type max ) const {
	return static_cast<size_type>( std::floor( (max - start_) / step_ ) );
}

wf::lg_generator::lg_generator( result_type start, result_type lstep, result_type gstep ): start_(start), t_(lstep), q_(result_type(1) + gstep / 2) {
	max_ = static_cast< size_type >( std::floor( result_type(2) / gstep - start / lstep ) );
}
wf::lg_generator::result_type wf::lg_generator::operator()( size_type i ) const {
	if( i <= max_ ) return ( start_ + t_ * i );
	return ( start_ + t_ * max_ ) * std::pow( q_, static_cast<int>(i - max_) ); // FIXME: pow(double,int) for <4.2.1
}
wf::lg_generator::size_type wf::lg_generator::size( result_type max ) const {
	result_type gmax = std::floor( std::log( max / ( start_ + t_ * max_ ) ) / std::log( q_ ) );
	if( gmax > result_type(0) ) return static_cast< size_type >( gmax ) + max_ + 1;
	return static_cast< size_type >(std::floor( (max - start_) / t_ ));
}
