/*
    $Id: weif.h 53 2009-11-01 18:03:28Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*--------------------------------------------------------------------------*/
/** @file weif.hxx
                  MASS project: TURBINA module file
                       header file for weif.c

 This module contains the collection of utilities suited to compute the
 altitude weighting functions for a given set of apertures for measurement
 of both normal and differential scintillation indices observed with
 a certain bandpass*energy distribution of incident light.

 \attention Spectral Energy Distributions must be photon numbers related, i.e.
 in [photons/\AA], not in [erg/\AA]. Meanwhile, they still are denoted as SED
 for simplicity.

 This module is based on the trial program wfm2.c written by A.Tokovinin
 and V. Kornilov (version Feb 26, 2001). The only bug found in that
 program was incorrect memory freeing by the (modified) free_vector() function
 taken from Numerical Recipes (NR) package.

 The module contains the definition of the weighting function class Weights
 and member functions for the weight calculation and input/output.
 The structure bears all the data and parameters which determine the weights
 and the respective altitude grid on which the weight matrix is computed.
 The parameters determine the z-grid, the number of altitude steps, the
 number of annular apertures with their sizes and effective wavelength.

 \b Usage

 Prior to any operation, the constructed weight structure should be filled by
 Weights::setaper(), Weights::setzgrid() and Weights::calcleff(). Being thus defined,
 then weight functions may be  either:
 - read from the chosen file : the Weights::checkfile() is used to to check the
 file  to coincide with the preset weight parameters (altitudes, wavelength,
 apertures). If the file fits the parameters, the weight function may be  read
 from the file with by Weights::read().
 - computed by Weights::calc().

 If one gets known the new spectral energy distribution, the need to recompute
 the weight for this SED may be verified by Weights::calcleff() with parameter \em
 fix==0. Also, the general checker/computer wf::calculation() is provided for
 the arbitrary input response function, SED and system geometry parameters.

 The parameters - the number of apertures, resulting weights (for apertures and
 their combinations) and number of altitude steps - can be accessed with
 Weights::getnaper(), Weights::getnw() and Weights::getnz(), respectively. The \em name
 of the weight is obtained by the Weights::name() function according to the column
 number in the weight matrix.

 Resulting weights may be shifted by altitude and/or recomputed for another
 altitude grid with Weights::interpolate(). Weight matrix can be written in the
 disk file by Weights::write() and released with Weights::clear(). Full release of
 memory in weight structure is done with Weights::setaper() with one zero
 parameter. The individual values of weight or altitude can be obtained with
 Weights::getval() and  Weights::getalt(); to copy some part of a certain weight
 function from the weight matrix in structure into a vector, use Weights::copy().

 The testing main() is active in module compiled as a separate executable (set
 WEIFTEST macro to do so). If WEIFNTEST is set zero, the executable shall be a
 simple weight computer given the responce and spectrum functions. If it is
 positive, the full circle of calculations and interpolations is done the
 specified number of times (say, for performance checks and memory leaks
 detection).

 @author N. Shatsky, Sternberg Institute (kolja@sai.msu.ru)

 @version 1.5: Polychromatic precise weight, in C++

*/
/* This header file is equipped with documentation made in comments in a
* format suitable fro production of LaTeX or HTML documentation with the
* doxygen software.
* Editor tabular size: 4                                                    */
/****************************************************************************/

#ifndef _WEIF_H
#define _WEIF_H

#include <string>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>

#include "vector_generator.h"

using namespace boost::numeric::ublas;

struct Spectral_data {
    std::string response;
    std::string weights;
    std::string spectra;
    Spectral_data( const std::string& r, const std::string& w, const std::string& s ) : response( r ), weights( w ), spectra( s ) {}
};

/****************************************************************************/
/*                     Weight function structure                            */
/****************************************************************************/

/**
This class contains all the data which describe the scintillation weighting
functions of altitudes for all entrance apertures with which it is measured.
The information on the apertures themselves and on the spectral characteristics
of the incident light is also present in members of this class.

*/

class Weights {
    unsigned int  mass_nchan;
    unsigned int  weights_number;
    vector<double> zgrid;           /**< vector of altitudes: z-grid [km] */
    matrix<double> ws;              /**< matrix of weights ws[aper][z] */
    vector<double> eps_inn;         /**< real inner diameters of apertures in [cm] */
    vector<double> eps_out;         /**< real outer diameters of apertures in [cm] */
    double leff;                    /**< effective wavelength of weight func. [mkm] */
    double lblue;                   /**< 50\% level wavelength at the blue side of SED [mkm] */
    double lred;                    /**< 50\% level wavelength at the red side of SED [mkm] */
public:
    Weights( int, const vector<double>& );
    ~Weights() {}

    const matrix<double>& weights() const { return ws; }
    const vector<double>& grid() const { return zgrid; }
    unsigned int apertures_number() const { return mass_nchan; }
    unsigned int weight_functions_number() const { return weights_number; }
    double lambda_eff() const { return leff ;}
    void interpolate( const Weights& refwf );
    void calculation( const Spectral_data&, const vector<double>&, const vector<double>& );

private:
/*    const char* name( int index ) const {
        static std::string abbr[] = { "A", "B", "C", "D", "AB", "AC", "AD", "BC", "BD", "CD" };
        assert( index < weights_number && index >= 0 );
        return abbr[index].c_str();
    }
*/
//    double getred() const { return lred ; }
//    double getblue() const { return lblue ; }

/**
	@brief	Writes weight matrix with parameters in disk ASCII file
	@param	filename	file name to write weight

	Procedure writes the matrix of weights which rows are preceded with the
	z-grid altitudes in the first column. The header is also created which
	contains all the parameters - the fields of Weights.

	The format of the header line is following:

	\verbatim
	# <KEY> <VALUE1> [<VALUE2> [VALUE3 VALUE4]]
	\endverbatim

	where <KEY> is a single-letter designation of parameter and <VALUE> - its
	value(s). Keys are listed below with names of respective Weights fields and
	description in brackets:

	\arg \c N --\em naper (number of apertures)
	\arg \c A -- \em D*eps_inn[0] \em D*eps_inn[0] (diameters of the 1st aperture)
	\arg \c B -- \em D*eps_inn[1] \em D*eps_inn[1] (diameters of the 2nd aperture)
	\arg ...
	\arg \c S -- \em leff \em lblue \em lred (effective wavelength and
	50\%-level points in [mkm])
	\arg \c Z -- \em z0 \em zmax \em dz \em dzmin (parameters for an altitude
	grid generation)
	\arg \c W --\em weightscale (the scale-factor for the weight values
	written in this file)

	After the header, the line goes which precedes the z-grid and wf-matrix
	values:

	\verbatim
	##Altit. W(A)/1E+11 W(B)/1E+11 ... W(AB)/1E+11
	\endverbatim
    std::cerr << "File " << filename << std::endl;
	where 1E+11 is a number to which the values of weights are scaled in a
	file (\em weightscale).

	Any similar line which starts with "##" is interpreted by Weights::read() as a
	comment.

	After the heading line "##Altit..." the z-grid and weight-matrix values go
	being separated with spaces	for each altitude; thus -- one line per each
	altitude.

*/

    void write( const std::string& ) const ;

/**
	@brief	Calculate the weight matrix for a given energy distribution
	@param	lambda	grid of wavelengths [mkm]
	@param	edist	incident energy distribution (given on lambda grid)
	@param	nedist	length of the \em lambda and \em edist grids

	Given the parameters in numeric members of the object (aperture geometry)
	and input spectrum, the weights are computed on a filled \em zgrid (see
	setzgrid()).

	The weights are computed for the polychromatic case of incident light SED
	which involves the Fourier Transform of the SED divided by the wavelength.
	This FT is made by wf::make_fft() once before calculation of all weights
	and is represented in two global arrays as cubic spline coefficients. In
	order to have smoother shape, the SED is shifted to its origin (starting
	wavelength \f$\lambda_{min}\f$) to dump the high-frequency oscillations in
	FT. The back-correction for this shift is made analytically after the
	spline interpolation for the given frequency.

	Before calculations, the weight	matrix is allocated IF it was NULL.
	If it was not NULL, it is assumed that it was already allocated (e.g. in
	a previous call of this routine, say, with different en.dist.).

	Then, for all apertures, their combinations and for all altitudes in the
	z-grid, the function wf::weight() is called to make the weight integration
	itself.

	At the end, the SED characteristics computer Weights::calcleff() is called to fill
	in the respective spectral fields. It appears that effective wavelength
	represents best the shape of weights if they are computed for the
	"quasi-gaussian" approximation of SED with its actual FWHM. Meanwhile, the
	deviations may be as high as 10\% from the exact shape, especially for
	differential indices of small apertures.

	The utility calls the external function \em progress() (if supplied as
	non-null) to visualize the degree of the made work. This degree, in
	percents, is an integer-type argument changing from 0 in the beginning to
	100 at the end. Second parameter is a message declaring what process is in
	progress. Progress function is called twice per altitude step.

	\b Performance

	At the IBM P-III 667 MHz machine, the calculations of weights with 26
	proportional steps of altitude (up to 20 km) and for 4 apertures take \b 14
	seconds.
*/
    void calc( vector<double>&, vector<double>& );

/**
	@brief	Calculate the effective wavelength and 50\%-levels of the light
		energy distribution and compares them with that set in member fields
	@param	lambda	grid of wavelengths [mkm]
	@param	edist	incident energy distribution (SED), given on \em lambda grid
	@param	nedist	number of nodes in \em lambda grid (points of en.dist.),
		or 0 to	reset the spectral information in weight structure
	@param	fix  	If \em fix==1: assign the computed spectral characteristics
		to Weights fields, or leave them otherwise (use wf::LEFFFIX or
		wf::LEFFTEST)
	@return int(SED characteristics differ from that set in the	structure) or
		the (negative) error code

	The effective wavelength and the wavelengths at which the relative light
	energy is equal to 50\% of the maximal are computed by the local linear
	interpolation for the given incident light SED. It is assumed that an array
	of relative intensities edist[] takes into account the instrument and
	atmospheric transmission functions. The computed effective wavelength and
	50\% points are first compared to the values in the fields \em leff, \em
	lblue and \em lred, and then assigned to them, if \em fix==1 and (!) the
	difference \em is significant.

	The result of comparison is returned: 0 if the difference is not
	significant, i.e. all three calculated numbers are within +/- \em
	EPSLEFF interval from the field values, and 1 if at least one of them
	differs by more than \em EPSLEFF. All wavelengths are in [mkm].

	If the parameter \nedist is supplied as 0, the calculations and comparison
	are skipped and the fields \em leff, \em lblue and \em lred are reset to
	zero. 0 is returned.

*/
    int calcleff( vector<double>& lambda, vector<double>& edist );

/**
	@brief	Interpolate the weights from the reference object to the z-grid
		set in the object with a given optional z-shift
	@param	refwf	Weights to get the weights from
	@param	shift	the shift of weight function along z-grid [km]
	@param	maxerr	pointer where to put the maximal weight error due to interp
		(works in POLINT option of compilation; may be null).
	@param	iz_maxerr	pointer to put the z-grid index of this error (may be
		null)
	@param	iw_maxerr	pointer to put the weight index of this error  (may be
		null)

	This function uses the NR recipe for interpolation of weight values from
	one z-grid into another. Input weight structure Weights must be fully
	determined; output structure must only contain the computed new z-grid: i.e.
	only the setzgrid() function must be called for it before interpolation.
	Other parameters must be empty - they will be assigned from the reference
	Weights fields.

	The parameter \em shift allows one to obtain the shifted weighting
	functions suited to work with the indices measured with the shifted pupil of
	the system (defocusing of the instrument). The \em negative altitudes
	when probed during the interpolation of the positively shifted functions are
	replaced with their absolute values. When negatively shifted function is
	stretched to reach the same \em zmax, the extrapolation case takes place.

	The rest parameters are three pointers to variables in which the user may
	receive the estimated maximal interpolation error of the weight, and the
	indices in z-grid and weight number where this maximal error was found. For
	example, for 4-aperture weight matrix (10 weights for each altitude),
	iz_maxerr = 2 and iw_maxerr = 5 will mean that maximal error \em maxerr was
	found at the third stage of the NEW altitude grid \em zgrid[2] and for the
	weight of AC-aperture combination.

*/

};                          // of the class Weights

/**
	@brief
	Additional to Weights, this namespace wf contains supplementary
	functions - getzshift() and the sample progress indicator progress(). Also,
	all the module constants are defined here.
*/

namespace wf {
/**
	@brief	Check the weight file to correspond to given system and spectral
		parameters. May recalculate on mismatch
	@param	sedfile	Stellar spectrum energy distribution file
	@param	response	System response curve file
	@param	wfile	Weight file to check and/or put the results of calculations
	@param	naper	Number of entrance apertures in scintillation device
   	@param  diam    see Weights::setaper()
   	@param  eps_inn see Weights::setaper()
   	@param  eps_out see Weights::setaper()
	@param	z0	altitude grid starting point (see Weights::setzgrid())
	@param	zmax	maximal altitude in weight matrix
	@param	dz	altitude step
	@param	dzmin	minimal step for propostional grid (0 for equidistant)
	@param	calcit	False: only check the file, True: recalculate if does not fit
	@param  progress	function to call to visualize the work stage (if needed)
	@result	the comparison result returned by Weights::checkfile() or error code

	This function makes use of all the functionalities presented by the class
	Weights to compute the weight functions set for a system of a given geometry
	and a certain spectrum of incident light (composed by the star's SED and the
	system response function).

	Via parameters, all the needed information on the geometry and spectrum is
	provided. The altitude grid, if left default (all values are 0), is taken
	as given by the parameters wf::Z0, wf::ZMAX, wf::DZ, wf::DZMIN.

	If some mismatch is found between the weight in file \a wfile and the
	configuration specified by the parameters, the result "false" is returned
	in case \a calcit=false.
	The particular reason of a mismatch is provided by the call to
	nr::erget(), nr::ermessage(). Don't forget to reset then the error before
	continuing with nr::erreset().

	If a mismatch is found and \a calcit is "true", the weights are recomputed
	in the file \a wfile. "True" is returned normally (the result of
	Weights::calc() actually).

	\note In case of recalculation, the "optimal" grid of altitudes specified by
	the parameters wf::Z0, wf::ZMAX, wf::DZ, wf::DZMIN is taken for calculations
	unless the mismatch between the file and parameters was indeed the grids
	mismatch. In latter case, the \a z0, \a zmax, \a dz and \a dzmin parameters
	are taken instead.
*/

/***************************************************************************/
/*                  Constant definitions                                   */
/***************************************************************************/
/** Writing weights to the file : the Scale factor to divide the values by */

const double WSCALE	= 1E11 ;

/**
    @brief  Generate the altitude grid given its parameters
    @param  z0  starting altitude [km]
    @param  zmax    altitude to reach [km]
    @param  dz  altitude modifier ([km] for equidistant grid, unitless for
    proportional)
    @param  dzmin   minimal altitude step [km]

    Compute the grid of altitude steps for calculation of weighting
    functions like follows:

    The grid may be either equidistant (if \em dzmin ==0) or
    proportional (\em dzmin !=0). The current restriction is
    that the altitude modifier  should be positive.

    Equidistant grid is given as:
    @f[
    z[i] = z0 + i * dz
    @f]
    where i is in [0,nz-1], nz conforms to the following condition:
    @f[
    z[0] + dz  (nz-2) < zmax
    @f]
    and
    @f[
    z[0] + dz  (nz-1) \ge zmax
    @f]

    Proportional grid is constructed as follows:
    \verbatim
    z[0] = z0
    z[i] = z[i-1] + step
    step0 = z[i-1] * dz
    step = ( dzmin > step0 ) ? (step = dzmin) : (step = step0)
    \endverbatim
    where i is in [0,nz-1], nz conforms to the following condition:
    @f[
    z[nz-2] < zmax
    @f]
    and
    @f[
    z[nz-1]  \ge zmax .
    @f]
    In both types of grid, z[nz-1] is replaced with \em zmax.

    This function is an interface to the lower-level private function zgrid_set()
    which works directly with the array \em Weights::zgrid.

    @see    setaper()
*/

/**
    @brief  Generate the grid of values given its generation parameters
    @param  zgrid   pointer double-type array
    @param  z0  starting value
    @param  zmax    value to reach
    @param  dz  value modifier
    @param  dzmin   minimal value step
    @return Number of steps in z, -6 if NOT NULL zgrid pointer, -3 if bad params

    This function implements the generation of a grid of abscissa values used
    by Weights::setzgrid() to generate the altitude grid in weight function
    structure. See weif.hxx for description of an algorithm given in
    wf::setzgrid() documentation.

    The function counts the steps that will be computed, allocates the
    double-type array (to which the first parameter must point) and fills in
    it. The array must be freed after usage (free_dvector()).

*/

    //template<class T> vector<T> grid_generate( unsigned int n, const T* z ) {
    template<class V> vector< typename V::value_type > grid_generate( const V& z ) {
        typedef V vector_type;
        typedef typename vector_type::value_type value_type;
        typedef typename vector_type::size_type size_type;
        typedef vector< value_type > return_vector_type;
        size_type n = z.size();
        const int z0 = 0;
        const int zmax = 1;
        const int dz = 2;
        const int dzmin = 3;
        return_vector_type h = scalar_vector<double>(1,z[z0]);
        if( n == 4 && z[zmax]> z[dz] && z[dzmin] != 0 ) {
            for( unsigned int i = 0; h(i) < z[zmax] ; i++ ) {
                h.resize(h.size()+1);
                h(i+1) = h(i) + std::max(0.5*h(i)*z[dz],z[dzmin]);
            }
        }
        if( n == 4 && z[zmax] > z[dz] && z[dzmin] == 0 ) {
            for( unsigned int i = 0; h(i) < z[zmax] ; i++ ) {
                h.resize(h.size()+1);
                h(i+1) = h(i) + z[dz];
            }
        }
        if( n > 1 && z[zmax]< z[dz] && z[dz] < z[dzmin] ) {
            h.resize(n);
            for( unsigned int i = 0; i < n ; i++ ) h(i) = z[i];
        }
        return h;
    }

    struct linear_generator {
        typedef double result_type;
        typedef size_t size_type;

        linear_generator( result_type start, result_type step );
        result_type operator()( size_type i ) const;
        size_type size( result_type max ) const;
    private:
        result_type start_,step_;
    };

    struct lg_generator {
        typedef double result_type;
        typedef size_t size_type;

        lg_generator( result_type start, result_type lstep, result_type gstep );
        result_type operator()( size_type i ) const;
        size_type size( result_type max ) const;
    private:
        result_type start_,t_,q_;
        size_type max_;
    };
};                   // end of namespace wf

#endif
